package com.example.ed448_goldilock

import androidx.annotation.NonNull
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import org.bouncycastle.crypto.Signer
import org.bouncycastle.crypto.generators.Ed448KeyPairGenerator
import org.bouncycastle.crypto.params.Ed448KeyGenerationParameters
import org.bouncycastle.crypto.params.Ed448PrivateKeyParameters
import org.bouncycastle.crypto.signers.Ed448Signer
import org.bouncycastle.math.ec.rfc8032.Ed448
import org.bouncycastle.util.encoders.Hex
import java.security.SecureRandom


/** Ed448GoldilockPlugin */
class Ed448GoldilockPlugin: FlutterPlugin, MethodCallHandler {

  /// The MethodChannel that will the communication between Flutter and native Android
  ///
  /// This local reference serves to register the plugin with the Flutter Engine and unregister it
  /// when the Flutter Engine is detached from the Activity
  private lateinit var channel : MethodChannel

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    channel = MethodChannel(flutterPluginBinding.binaryMessenger, "ed448_goldilock")
    channel.setMethodCallHandler(this)
  }

  override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
    if (call.method == "generatePrivKey") {
      val RANDOM = SecureRandom();
      val kpg =  Ed448KeyPairGenerator()
      kpg.init(Ed448KeyGenerationParameters(RANDOM))
      val kp = kpg.generateKeyPair()
      val privateKey = kp.private as Ed448PrivateKeyParameters
      result.success(privateKey.encoded.toHexString())
    }
    else if (call.method == "getPublicKey") {
      val key = call.argument<String>("key");
      val privateKey = Ed448PrivateKeyParameters(
              key?.decodeHex(), 0);
      val publicKey = privateKey.generatePublicKey();

      val pubToReturn =publicKey.encoded.toHexString()
      result.success(pubToReturn)
    }

    else if (call.method == "sign") {
      val key = call.argument<String>("key");
      val msg = call.argument<String>("msg");

      val msgBytes =  msg?.decodeHex();
      val privateKey = Ed448PrivateKeyParameters(
              key?.decodeHex(), 0);
      val signBytes = ByteArray(Ed448.SIGNATURE_SIZE);
       if (msgBytes!=null){
  privateKey.sign(Ed448.Algorithm.Ed448 , ByteArray(0), msgBytes, 0, msgBytes.size,signBytes,0);
}
      result.success(signBytes.toHexString())
    }
    else if (call.method == "concatenatedPubKeyWithSign") {
      val key = call.argument<String>("key");
      val msg = call.argument<String>("msg");

      val msgBytes =  msg?.decodeHex();
      val privateKey = Ed448PrivateKeyParameters(
              key?.decodeHex(), 0);
      val signBytes = ByteArray(Ed448.SIGNATURE_SIZE);
      if (msgBytes!=null){
        privateKey.sign(Ed448.Algorithm.Ed448 , ByteArray(0), msgBytes, 0, msgBytes.size,signBytes,0);
      }
      val pubKey = privateKey.generatePublicKey().encoded;
      val finalResult = signBytes + pubKey
      result.success(finalResult.toHexString());
    }
    else if (call.method == "verifySignature") {
      val msgHash = call.argument<String>("msgHash");
      val signedMsg = call.argument<String>("signedMsg");
      val pubKey = call.argument<String>("pubKey");

      val msgBytes =  msgHash?.decodeHex();
      val signedBytes = signedMsg?.decodeHex()
      val pubBytes = pubKey?.decodeHex();
      var res = false;
      if (msgBytes!=null){
        res = Ed448.verify(signedBytes,0,pubBytes,0,ByteArray(0),msgBytes,0,msgBytes.size);
      }
      println(res)
      result.success(res);

    }
    else {
      result.notImplemented()
    }
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    channel.setMethodCallHandler(null)
  }
  fun getName(): String? {
    return "Ed448"
  }
  private fun createSigner(context: ByteArray): Signer? {
    return  Ed448Signer(context);
  }
  fun ByteArray.toHexString() = joinToString("") { "%02x".format(it) }
  fun String.decodeHex(): ByteArray = chunked(2)
          .map { it.toInt(16).toByte() }
          .toByteArray()
}
